use std::fmt::Debug;

use futures::channel::mpsc::{channel, Receiver};
use futures_util::SinkExt;
use prost::DecodeError;
use prost_reflect::{FileDescriptor, ReflectMessage};
use prost_types::{FileDescriptorProto};
use tonic::{async_trait, Request, Status};
use tonic::transport::Channel;
use tracing::{error, trace};

use crate::client::proto_db_client::ProtoDbClient;
use crate::event::*;
use crate::txn::*;
use crate::types::AnyExt;
use crate::Error::Empty;
use crate::{Result};
use crate::{GetOptions, ProtoDb, Reader, Registerer, SetOptions, Transactioner, WatchOptions, Watcher, Writer};

tonic::include_proto!("linka.cloud.protodb");

static PROTODB_READ_ONLY: &str = "protodb-read-only";

#[derive(Debug, Clone)]
pub struct Client {
    inner: ProtoDbClient<Channel>,
}

impl Client {
    pub fn new(inner: Channel) -> Self {
        let inner = ProtoDbClient::new(inner);
        Client { inner }
    }

    async fn do_txn(&mut self, ro: bool) -> Result<Txn> {
        trace!("create transaction");
        let (req_tx, req_rx) = channel::<TxRequest>(1);
        let (mut res_tx, res_rx) = channel::<Result<Option<TxResponse>, Status>>(1);

        let mut req = Request::new(req_rx);
        if ro {
            req.metadata_mut().insert(PROTODB_READ_ONLY, "true".parse().expect("failed to parse metadata"));
        }
        let mut s = self.inner.tx(req).await?.into_inner();
        tokio::spawn(async move {
            loop {
                match s.message().await {
                    Ok(res) => {
                        if let Err(err) = res_tx.send(Ok(res)).await {
                            error!(error = ?err, "Failed to forward response");
                            return;
                        }
                    }
                    Err(err) => {
                        if let Err(err) = res_tx.send(Err(err)).await {
                            if err.is_disconnected() {
                                trace!("disconnected");
                                return;
                            }
                            error!(error = ?err, "Failed to forward response");
                        }
                        return;
                    }
                };
            }
        });
        Ok(Txn::new(req_tx, res_rx))
    }
}

impl ProtoDb<Txn> for Client {}

#[async_trait]
impl Reader for Client {
    #[tracing::instrument]
    async fn get<M: ReflectMessage + Default>(&mut self, search: &M) -> Result<(Vec<M>, Option<PagingInfo>)> {
        self.get_opts(search, GetOptions::default()).await
    }

    #[tracing::instrument]
    async fn get_opts<M: ReflectMessage + Default>(&mut self, search: &M, opts: GetOptions) -> Result<(Vec<M>, Option<PagingInfo>)> {
        trace!(search = ?search);
        let GetOptions { filter, paging, field_mask } = opts;
        let GetResponse { results, paging } = self.inner.get(
            Request::new(GetRequest {
                search: Some(search.to_any()),
                filter,
                paging,
                field_mask,
            })).await?.into_inner();
        Ok((results.iter().map(|any| M::from_any(any)).collect::<Result<Vec<M>, DecodeError>>()?, paging))
    }
}

#[async_trait]
impl Writer for Client {
    #[tracing::instrument]
    async fn set<M: ReflectMessage + Default>(&mut self, m: &M) -> Result<M> {
        self.set_opts(m, SetOptions::default()).await
    }

    #[tracing::instrument]
    async fn set_opts<M: ReflectMessage + Default>(&mut self, m: &M, opts: SetOptions) -> Result<M> {
        trace!(set = ?m);
        let SetOptions { field_mask, ttl } = opts;
        let SetResponse { result } = self.inner.set(Request::new(SetRequest {
            payload: Some(m.to_any()),
            field_mask,
            ttl,
        })).await?.into_inner();
        Ok(M::from_any(&result.ok_or(Empty)?)?)
    }

    #[tracing::instrument]
    async fn delete<M: ReflectMessage + Default>(&mut self, m: &M) -> Result<()> {
        trace!(delete = ?m);
        self.inner.delete(Request::new(DeleteRequest { payload: Some(m.to_any()) })).await?;
        Ok(())
    }
}

#[async_trait]
impl Watcher for Client {
    #[tracing::instrument]
    async fn watch<M: ReflectMessage + Default + 'static>(&mut self, m: &M) -> Result<Receiver<Result<Event<M>, DecodeError>>> {
        self.watch_opts(m, WatchOptions::default()).await
    }

    #[tracing::instrument]
    async fn watch_opts<M: ReflectMessage + Default + 'static>(&mut self, m: &M, opts: WatchOptions) -> Result<Receiver<Result<Event<M>, DecodeError>>> {
        trace!(watch = ?m);
        let WatchOptions { filter } = opts;
        let mut s = self.inner.watch(Request::new(WatchRequest {
            search: Some(m.to_any()),
            filter,
        })).await?.into_inner();
        let (mut tx, rx) = channel::<Result<Event<M>, DecodeError>>(1);
        tokio::spawn(async move {
            trace!(method = "watch", "watching ...");
            while let Ok(Some(res)) = s.message().await {
                // ignore unknown events
                // it is expected to receive an empty event when the watch is first created
                // this is a hack implemented server side to allow us to get a stream
                if res.r#type == 0 {
                    continue;
                }
                if let Err(err) = tx.send(Event::from_any(res)).await {
                    if err.is_disconnected() {
                        trace!("disconnected");
                        return;
                    }
                    error!(method = "watch", error = ?err, "failed to forward event");
                    return;
                }
            }
        });
        Ok(rx)
    }
}

#[async_trait]
impl Transactioner<Txn> for Client {
    #[tracing::instrument]
    async fn txn(&mut self) -> Result<Txn> {
        self.do_txn(false).await
    }

    #[tracing::instrument]
    async fn read_only_txn(&mut self) -> Result<Txn> {
        self.do_txn(true).await
    }
}

#[async_trait]
impl Registerer for Client {
    #[tracing::instrument]
    async fn register<M: ReflectMessage>(&mut self, m: &M) -> Result<()> {
        self.register_file(&m.descriptor().parent_file()).await
    }

    #[tracing::instrument]
    async fn register_file(&mut self, fd: &FileDescriptor) -> Result<()> {
        self.register_file_proto(fd.file_descriptor_proto()).await
    }

    #[tracing::instrument]
    async fn register_file_proto(&mut self, fdp: &FileDescriptorProto) -> Result<()> {
        trace!(proto = fdp.name, "register_file_proto");
        self.inner.register(Request::new(RegisterRequest { file: Some(fdp.clone()) })).await?;
        Ok(())
    }
}
