#![deny(warnings)]
#![warn(unused_extern_crates)]
#![deny(clippy::todo)]
#![deny(clippy::unimplemented)]
#![warn(clippy::unwrap_used)]
#![deny(clippy::expect_used)]
#![deny(clippy::panic)]
#![deny(clippy::unreachable)]
#![deny(clippy::await_holding_lock)]
#![deny(clippy::needless_pass_by_value)]
#![deny(clippy::trivially_copy_pass_by_ref)]

use crate::client::PagingInfo;
pub use crate::event::Event;
pub use client::{Client};
pub use event::Type;
use futures::channel::mpsc::{Receiver, SendError};
pub use options::*;
use prost::DecodeError;
use prost_reflect::{FileDescriptor, ReflectMessage};
use prost_types::FileDescriptorProto;
pub use protofilters::Where;
use tonic::{async_trait, Status};
pub use txn::*;

mod client;
mod txn;
pub mod event;

pub mod protofilters;
pub mod types;
mod options;

pub type Result<T, E = Error> = std::result::Result<T, E>;

#[derive(thiserror::Error, Debug)]
pub enum Error {
    #[error("decode error: {0}")]
    Decode(#[from] DecodeError),
    #[error("status error: {0}")]
    Status(#[from] Status),
    #[error("send error: {0}")]
    Send(#[from] SendError),
    #[error("empty response")]
    Empty,
}

pub trait ProtoDb<T: Transaction>: Reader + Writer + Watcher + Transactioner<T> + Registerer {}
pub trait Transaction: Reader + Writer + Committer {}

#[async_trait]
pub trait Reader {
    async fn get<M: ReflectMessage + Default>(&mut self, search: &M) -> Result<(Vec<M>, Option<PagingInfo>)>;
    async fn get_opts<M: ReflectMessage + Default>(&mut self, search: &M, opts: GetOptions) -> Result<(Vec<M>, Option<PagingInfo>)>;
}

#[async_trait]
pub trait Writer {
    async fn set<M: ReflectMessage + Default>(&mut self, m: &M) -> Result<M>;
    async fn set_opts<M: ReflectMessage + Default>(&mut self, m: &M, opts: SetOptions) -> Result<M>;
    async fn delete<M: ReflectMessage + Default>(&mut self, m: &M) -> Result<()>;
}

#[async_trait]
pub trait Watcher {
    async fn watch<M: ReflectMessage + Default + 'static>(&mut self, m: &M) -> Result<Receiver<Result<Event<M>, DecodeError>>>;
    async fn watch_opts<M: ReflectMessage + Default + 'static>(&mut self, m: &M, opts: WatchOptions) -> Result<Receiver<Result<Event<M>, DecodeError>>>;
}

#[async_trait]
pub trait Registerer {
    async fn register<M: ReflectMessage>(&mut self, m: &M) -> Result<()>;
    async fn register_file(&mut self, fd: &FileDescriptor) -> Result<()>;
    async fn register_file_proto(&mut self, fdp: &FileDescriptorProto) -> Result<()>;
}

#[async_trait]
pub trait Transactioner<T: Transaction> {
    async fn txn(&mut self) -> Result<T>;
    async fn read_only_txn(&mut self) -> Result<T>;
}

#[async_trait]
pub trait Committer {
    async fn commit(&mut self) -> Result<()>;
}
