use prost_types::Timestamp;
use crate::protofilters::filter::Match;

tonic::include_proto!("linka.cloud.protofilters");

pub use crate::client::Paging;
pub use crate::client::PagingInfo;

#[derive(Clone,Debug)]
pub struct Where {
    expr: Expression,
    not: bool,
}

impl Into<Expression> for Where {
    fn into(self) -> Expression {
        self.expr
    }
}

impl Where {
    pub fn new(field: &str) -> Self {
        Where {
            expr: Expression {
                condition: Some(FieldFilter {
                    field: field.to_string(),
                    ..Default::default()
                }),
                ..Expression::default()
            },
            not: false,
        }
    }

    pub fn and(mut self, other: Where) -> Where {
        self.expr.and_exprs.push(other.expr);
        self
    }

    pub fn or(mut self, other: Where) -> Where {
        self.expr.or_exprs.push(other.expr);
        self
    }

    pub fn not(mut self) -> Where {
        self.not = true;
        self
    }

    pub fn string(self) -> String {
        String {
            w: self,
            insensitive: false,
        }
    }

    pub fn number(self) -> Number {
        Number {
            w: self,
        }
    }

    pub fn bool(self) -> Bool {
        Bool {
            w: self,
        }
    }

    pub fn null(self) -> Null {
        Null {
            w: self,
        }
    }

    pub fn time(self) -> Time {
        Time {
            w: self,
        }
    }

    pub fn duration(self) -> Duration {
        Duration {
            w: self,
        }
    }
}

pub struct String {
    w: Where,
    insensitive: bool,
}

impl String {
    pub fn not(mut self) -> String {
        self.w.not = true;
        self
    }

    pub fn insensitive(mut self) -> String {
        self.insensitive = true;
        self
    }

    pub fn eq(self, value: &str) -> Where {
        self.build(string_filter::Condition::Equals(value.to_string()))
    }

    pub fn has_prefix(self, value: &str) -> Where {
        self.build(string_filter::Condition::HasPrefix(value.to_string()))
    }

    pub fn has_suffix(self, value: &str) -> Where {
        self.build(string_filter::Condition::HasSuffix(value.to_string()))
    }

    pub fn regex(self, value: &str) -> Where {
        self.build(string_filter::Condition::Regex(value.to_string()))
    }

    pub fn in_values(self, value: &[&str]) -> Where {
        self.build(string_filter::Condition::In(string_filter::In {
            values: value.iter().map(|s| s.to_string()).collect(),
        }))
    }

    fn build(mut self, condition: string_filter::Condition) -> Where {
        self.w.expr.condition.as_mut().unwrap().filter = Some(Filter {
            r#match: Some(Match::String(StringFilter {
                condition: Some(condition),
                case_insensitive: self.insensitive,
            })),
            not: self.w.not,
        });
        self.w
    }
}

pub struct Number {
    w: Where,
}

impl Number {
    pub fn not(mut self) -> Number {
        self.w.not = true;
        self
    }

    fn build(mut self, filter: number_filter::Condition) -> Where {
        self.w.expr.condition.as_mut().unwrap().filter = Some(Filter {
            r#match: Some(Match::Number(NumberFilter {
                condition: Some(filter),
            })),
            not: self.w.not,
        });
        self.w
    }

    pub fn eq(self, value: f64) -> Where {
        self.build(number_filter::Condition::Equals(value))
    }

    pub fn gt(self, value: f64) -> Where {
        self.build(number_filter::Condition::Sup(value))
    }

    pub fn lt(self, value: f64) -> Where {
        self.build(number_filter::Condition::Inf(value))
    }

    pub fn in_values(self, value: &[f64]) -> Where {
        self.build(number_filter::Condition::In(number_filter::In {
            values: value.iter().map(|s| *s).collect(),
        }))
    }
}

pub struct Bool {
    w: Where,
}

impl Bool {
    fn build(mut self, b: bool) -> Where {
        self.w.expr.condition.as_mut().unwrap().filter = Some(Filter {
            r#match: Some(Match::Bool(BoolFilter {
                equals: b,
            })),
            not: self.w.not,
        });
        self.w
    }
    pub fn not(mut self) -> Bool {
        self.w.not = true;
        self
    }
    #[allow(clippy::wrong_self_convention)]
    pub fn is_true(self) -> Where {
        self.build(true)
    }
    #[allow(clippy::wrong_self_convention)]
    pub fn is_false(self) -> Where {
        self.build(false)
    }
}

pub struct Null {
    w: Where,
}

impl Null {
    pub fn not(mut self) -> Null {
        self.w.not = true;
        self
    }
    #[allow(clippy::wrong_self_convention)]
    pub fn is_null(mut self) -> Where {
        self.w.expr.condition.as_mut().unwrap().filter = Some(Filter {
            r#match: Some(Match::Null(NullFilter {})),
            not: self.w.not,
        });
        self.w
    }
}

pub struct Time {
    w: Where,
}

impl Time {
    pub fn not(mut self) -> Time {
        self.w.not = true;
        self
    }
    fn build(mut self, filter: time_filter::Condition) -> Where {
        self.w.expr.condition.as_mut().unwrap().filter = Some(Filter {
            r#match: Some(Match::Time(TimeFilter {
                condition: Some(filter),
            })),
            not: self.w.not,
        });
        self.w
    }
    pub fn eq(self, value: Timestamp) -> Where {
        self.build(time_filter::Condition::Equals(value))
    }
    pub fn before(self, value: Timestamp) -> Where {
        self.build(time_filter::Condition::Before(value))
    }
    pub fn after(self, value: Timestamp) -> Where {
        self.build(time_filter::Condition::After(value))
    }
}

pub struct Duration {
    w: Where,
}

impl Duration {
    pub fn not(mut self) -> Duration {
        self.w.not = true;
        self
    }
    fn build(mut self, filter: duration_filter::Condition) -> Where {
        self.w.expr.condition.as_mut().unwrap().filter = Some(Filter {
            r#match: Some(Match::Duration(DurationFilter {
                condition: Some(filter),
            })),
            not: self.w.not,
        });
        self.w
    }
    pub fn eq(self, value: prost_types::Duration) -> Where {
        self.build(duration_filter::Condition::Equals(value))
    }
    pub fn gt(self, value: prost_types::Duration) -> Where {
        self.build(duration_filter::Condition::Sup(value))
    }
    pub fn lt(self, value: prost_types::Duration) -> Where {
        self.build(duration_filter::Condition::Inf(value))
    }
}
