use futures::channel::mpsc::{Receiver, Sender};
use futures_util::SinkExt;
use prost_reflect::ReflectMessage;
use tokio_stream::StreamExt;
use tonic::{async_trait, Status};
use tracing::{error, trace};

use crate::client::*;
use crate::types::*;
use crate::{Committer, GetOptions, Reader, SetOptions, Transaction, Writer, Result, Error};

#[derive(Debug)]
pub struct Txn {
    tx: Sender<TxRequest>,
    rx: Receiver<Result<Option<TxResponse>, Status>>,
}

impl Txn {
    pub(crate) fn new(tx: Sender<TxRequest>, rx: Receiver<Result<Option<TxResponse>, Status>>) -> Self {
        Txn { tx, rx }
    }
}

impl Transaction for Txn {}

#[async_trait]
impl Reader for Txn {
    #[tracing::instrument]
    async fn get<M: ReflectMessage + Default>(&mut self, search: &M) -> Result<(Vec<M>, Option<PagingInfo>)> {
        self.get_opts(search, GetOptions::default()).await
    }

    #[tracing::instrument]
    async fn get_opts<M: ReflectMessage + Default>(&mut self, search: &M, opts: GetOptions) -> Result<(Vec<M>, Option<PagingInfo>)> {
        trace!(search = ?search);
        let GetOptions { filter, paging, field_mask } = opts;
        let req = GetRequest { search: Some(search.to_any()), filter, paging, field_mask };
        let res = self.tx.send(TxRequest { request: Some(tx_request::Request::Get(req)) }).await;
        if let Err(err) = res {
            error!(error = ?err);
            return Err(err.into());
        }
        let res = self.rx.next().await.ok_or(Error::Empty)?;
        match res {
            Ok(Some(TxResponse { response: Some(tx_response::Response::Get(res)) })) => {
                let mut v = Vec::new();
                for any in res.results.iter() {
                    v.push(M::from_any(any)?);
                }
                Ok((v, res.paging))
            }
            Err(err) => Err(Error::Status(err)),
            _ => Err(Error::Empty),
        }
    }
}

#[async_trait]
impl Writer for Txn {
    #[tracing::instrument]
    async fn set<M: ReflectMessage + Default>(&mut self, m: &M) -> Result<M> {
        self.set_opts(m, SetOptions::default()).await
    }

    #[tracing::instrument]
    async fn set_opts<M: ReflectMessage + Default>(&mut self, m: &M, opts: SetOptions) -> Result<M> {
        trace!(message = ?m);
        let SetOptions { ttl, field_mask } = opts;
        let req = SetRequest { payload: Some(m.to_any()), field_mask, ttl };
        let res = self.tx.send(TxRequest { request: Some(tx_request::Request::Set(req)) }).await;
        if let Err(err) = res {
            error!(error = ?err);
            return Err(Error::Send(err));
        }
        let res = self.rx.next().await.ok_or(Error::Empty)?;
        match res {
            Ok(Some(TxResponse { response: Some(tx_response::Response::Set(res)) })) => {
                let m = M::from_any(&res.result.ok_or(Error::Empty)?)?;
                Ok(m)
            }
            Err(err) => Err(Error::Status(err)),
            _ => Err(Error::Empty),
        }
    }

    #[tracing::instrument]
    async fn delete<M: ReflectMessage + Default>(&mut self, m: &M) -> Result<()> {
        trace!(message = ?m);
        let req = DeleteRequest { payload: Some(m.to_any()) };
        let res = self.tx.send(TxRequest { request: Some(tx_request::Request::Delete(req)) }).await;
        if let Err(err) = res {
            error!(error = ?err);
            return Err(Error::Send(err));
        }
        let res = self.rx.next().await.ok_or(Error::Empty)?;
        match res {
            Ok(Some(TxResponse { response: Some(tx_response::Response::Delete(_)) })) => {
                Ok(())
            }
            Err(err) => Err(Error::Status(err)),
            _ => Err(Error::Empty),
        }
    }
}

#[async_trait]
impl Committer for Txn {
    #[tracing::instrument]
    async fn commit(&mut self) -> Result<()> {
        trace!("committing transaction");
        let res = self.tx.send(TxRequest { request: Some(tx_request::Request::Commit(true)) }).await;
        if let Err(err) = res {
            error!(error = ?err);
            return Err(Error::Send(err));
        }
        let res = self.rx.next().await.ok_or(Error::Empty)?;
        match res {
            Ok(Some(TxResponse { response: Some(tx_response::Response::Commit(_)) })) => {
                Ok(())
            }
            Err(err) => Err(Error::Status(err)),
            _ => Err(Error::Empty),
        }
    }
}
